require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/cast/cast_level'
require 'uu/os/cast/cast_create'
require 'uu/os/cast/cast_attributes'
require 'uu/os/cast/cast_get_cast_list'
require 'uu/os/cast/cast_get_access_role_cast_list'
require 'uu/os/cast/cast_access_role_attributes'
require 'uu/os/cast/cast_set_attributes'
require 'uu/os/cast/cast_get_attributes'
require 'uu/os/cast/cast_delete'
require 'uu/os/cast/cast_state'
require 'uu/os/cast/cast_exists'
require 'uu/os/cast/cast_verify_cast_existence'

# Module Unicorn Universe.
module UU
  # Module Operating System.
  module OS
    # Module Cast.
    module Cast
      extend self

      # Service path
      PATH = 'ues/core/role/UESCast'
      # Casts a subject (role or group) to another subject (role or group). Only roles and
      # groups from the same territory are supported at the moment (therefore it is not
      # possible to cast across territories and form pseudo business access role relation
      # at the moment). Cast is not an entity, it is a relation between two subjects (and these
      # two subjects identify the relation). Note that only some combinations
      # of subjects in a cast relation are permitted.
      #
      # @param [String, UU::OS::UESURI] subject_uri UESURI of the subject into which another subject is casted
      # @param [CastCreate] cast DTO containing attributes of the cast
      # @option cast [String, UU::OS::UESURI] :casted_subject_uri UESURI of the subject (role or group) which is casted to the subject identified by the command main entity. Cannot be nil.
      # @option cast [String, UU::OS::Cast::CastLevel] :level The level of the new cast. The attribute cannot be set at the same time as the {#executive executive} attribute. If set to nil, {UU::OS::Cast::CastLevel::EXECUTIVE EXECUTIVE} level will be used.
      # @option cast [Boolean] :primary Specifies whether the cast will be the primary cast to the role. Applicable only if the subject_uri (main entity) is a role. If set to true, the cast must be {#visible visible} in {UU::OS::Cast::CastLevel::EXECUTIVE EXECUTIVE} {#level level}. If set to nil, the cast will not become primary unless it is the first cast to the subject_uri (main entity).
      # @option cast [String] :date_to The date when the cast will be automatically set to the {UU::OS::Cast::CastState::PASSIVE PASSIVE} state. If the subject is a Role and the cast is the {UU::OS::Cast::CastCreate#primary primary} cast to the Role, then there must be either another {UU::OS::Cast::CastState::ACTIVE ACTIVE} visible cast with {UU::OS::Cast::CastLevel::EXECUTIVE EXECUTIVE} {#level level} to the Role which will become the primary cast, or no cast at all; otherwise the cast state will not be changed.
      # @option cast [Boolean] :visible Specifies whether the cast is visible in Show Casts use case. If set to nil, the cast will be set as visible. If set to false, the cast must not be set as {#primary primary}.
      #
      # @example
      #   # Cast a role in a group
      #   UU::OS::Cast.create('ues:TERRITORY:GROUP',
      #           :casted_subject_uri => 'ues:TERRITORY:ROLE'
      #   )
      #
      #   # Create a primary cast of an access role in a role
      #   UU::OS::Cast.create('ues:TERRITORY:ROLE',
      #           :casted_subject_uri => 'ues:TERRITORY:ACCESS_ROLE',
      #           :primary => true
      #   )
      #
      #   # Cast role 1 in role 2 in assistance level
      #   UU::OS::Cast.create('ues:TERRITORY:ROLE2',
      #           :casted_subject_uri => 'ues:TERRITORY:ROLE1',
      #           :level => 'ASSISTANCE'
      #   )
      def create(subject_uri, cast = nil)
        svc = UU::OS::REST::RemoteClient.new(Cast)
        payload = UU::OS::Cast::CastCreate.new(cast).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('create', subject_uri, payload)
        end
      end

      # Returns attributes of one cast. Cast is identified by two subjects.
      # The main entity (a subject to which the other is casted) and parameter casted_subject_uri
      # in the DTO (the casted subject).
      #
      # @param [String, UU::OS::UESURI] subject_uri UESURI of the subject to which the other is casted
      # @param [CastGetAttributes] dto DTO containing UESURI of the casted subject
      # @option dto [String, UU::OS::UESURI] :casted_subject_uri UESURI of the subject (role or group) which is casted to the subject identified by the command main entity. Cannot be nil.
      #
      # @return [CastAttributes] DTO containing the cast attributes
      #
      # @example
      #   # Get attributes of a cast (ROLE2 is casted in ROLE1)
      #   attributes = UU::OS::Cast.get_attributes('ues:TERRITORY:ROLE1',
      #           :casted_subject_uri => 'ues:TERRITORY:ROLE2'
      #   )
      def get_attributes(subject_uri, dto)
        svc = UU::OS::REST::RemoteClient.new(Cast)
        get_dto = UU::OS::Cast::CastGetAttributes.new(dto)
        svc.add_parameter('castedSubjectUri', get_dto.casted_subject_uri)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttributes', subject_uri)
          return Cast::CastAttributes.new(res)
        end
      end

      # Returns list of casts to the specified subject (role or group). The list
      # is sorted by the date of creation of a cast by default and due to backwards compatibility
      # are returned only casts with {CastDirection::IN IN} cast direction by default.
      #
      # @param [String, UU::OS::UESURI] subject_uri UESURI of the subject whose casts will be listed.
      # @param [CastGetCastList] criteria Specific options.
      # @option criteria [String] :query Query for filtering and ordering of the result list (more information in {https://uu.unicornuniverse.eu/ues/sesm?&SessFree=ues:UNI-BT:UES.DOC%2FCMD_V1%2FUESQUERY uuQuery - Guideline}). It is possible to filter the result list by castedSubjectName, castedSubjectCode, dateFrom, state, level, primary and castDirection.
      #
      # @return [Array] An Array with the command result. Contains instances of {CastAttributes}.
      #
      # @example
      #   # Returns the primary cast in the role ROLE
      #   UU::OS::Cast.get_cast_list('ues:TERRITORY:ROLE', :query => 'castDirection=IN AND primary=true')
      #
      #   # Returns all executive visible casts of an access role in other roles
      #   UU::OS::Cast.get_cast_list('ues:TERRITORY:ACCESS_ROLE', :query => 'castDirection=TO AND state=ACTIVE AND level=EXECUTIVE')
      def get_cast_list(subject_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(Cast)
        dto = Cast::CastGetCastList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getCastList', subject_uri)
          return UU::OS::REST::ResultList.new(Cast::CastGetCastList, Cast::CastAttributes, res)
        end
      end

      # Returns list of access roles (personal access roles and uuEEs) casted directly or indirectly to the specified
      # subject (role or group). The list is sorted by the casted subject name and subsequently by casted subject code
      # by default.
      #
      # @param [String, UU::OS::UESURI] subject_uri UESURI of the subject into which the personal access roles are casted.
      # @param [CastGetCastList] criteria Object containing paging and filtering criteria.
      # @option criteria [String] :query Query for filtering and ordering of the result list (more information in {https://uu.unicornuniverse.eu/ues/sesm?&SessFree=ues:UNI-BT:UES.DOC%2FCMD_V1%2FUESQUERY uuQuery - Guideline}). It is possible to filter the result list by castedSubjectName and castedSubjectCode.
      #
      # @return [Array] An Array with the command result. Contains instances of {Cast::CastAccessRoleAttributes}.
      #
      # @example
      #   # Returns ordered list of casted access roles whose name starts with 'Name Prefix' and code with 'Code Prefix'
      #   UU::OS::Cast.get_access_role_cast_list('ues:TERRITORY:ROLE_OR_GROUP',
      #         :query => "castedSubjectName='Name Prefix%' AND castedSubjectCode='Code Prefix%' ORDER BY castedSubjectName")
      def get_access_role_cast_list(subject_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(Cast)
        dto = Cast::CastGetAccessRoleCastList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAccessRoleCastList', subject_uri)
          return UU::OS::REST::ResultList.new(Cast::CastGetAccessRoleCastList, Cast::CastAccessRoleAttributes, res)
        end
      end

      # Sets basic attributes of the cast. Cast is not an entity, it is a relation between
      # two subjects (and these two subjects identify the relation).
      #
      # @param [String, UU::OS::UESURI] subject_uri UESURI of the subject into which another subject is casted
      # @param [CastSetAttributes] cast DTO containing attributes of the cast
      # @option cast [String, UU::OS::UESURI] :casted_subject_uri UESURI of the subject (role or group) which is casted to the subject identified by the command main entity. Cannot be nil.
      # @option cast [String, UU::OS::Cast::CastLevel] :level The level of the cast. The attribute cannot be set at the same time as the {#executive executive} attribute. If the level is changed, the value of the {#executive executive} attribute will be changed as well (true for the {UU::OS::Cast::CastLevel::EXECUTIVE EXECUTIVE} level, false for the {UU::OS::Cast::CastLevel::ASSISTENCE ASSISTENCE} or {UU::OS::Cast::CastLevel::GUEST GUEST} level).
      # @option cast [Boolean] :primary Specifies whether the cast will be the primary cast to the role. Applicable only if the subject_uri (main entity) is a role. If set to true, the cast must be {#visible visible} in {UU::OS::Cast::CastState::ACTIVE ACTIVE} {#state state} and {UU::OS::Cast::CastLevel::EXECUTIVE EXECUTIVE} {#level level}. If the cast is primary and it is set to false, another cast (the oldest {#visible visible} cast in {UU::OS::Cast::CastState::ACTIVE ACTIVE} {#state state} and {UU::OS::Cast::CastLevel::EXECUTIVE EXECUTIVE} {#level level}) to the subject_uri (main entity) will become the primary cast. If set to nil, the primary cast will not be changed.
      # @option cast [String] :date_to The date when the cast will be automatically set to the {UU::OS::Cast::CastState::PASSIVE PASSIVE} {#state state}. If the subject is a Role and the cast is the {#primary primary} cast to the Role, then there must be either another {UU::OS::Cast::CastState::ACTIVE ACTIVE} {#visible visible} cast with {UU::OS::Cast::CastLevel::EXECUTIVE EXECUTIVE} {#level level} to the Role which will become the primary cast, or no cast at all; otherwise the cast state will not be changed. If set to an empty string, the date will be erased. If set to nil, the value will not be modified.
      # @option cast [String, UU::OS::Cast::CastState] :state Deprecated. In territories with cast strategies of 2 or higher, all casts are either ACTIVE or they do not exist at all, so if you set a cast into a PASSIVE state, the cast will be permanently deleted. In territories with a cast strategy of 1 or lower, PASSIVE casts are just ignored while only ACTIVE casts are used in cast calculations.
      # @option cast [Boolean] :visible Specifies whether the cast is visible in Show Casts use case. If set to nil, the value will not be modified. If set to false, the cast must not be set as {#primary primary}.
      #
      # @example
      #   # Change an existing cast to guest
      #   UU::OS::Cast.set_attributes('ues:TERRITORY:ROLE2',
      #           :casted_subject_uri => 'ues:TERRITORY:ROLE1',
      #           :level => UU::OS::Cast::CastLevel::GUEST
      #   )
      #
      #   # Change an existing cast of role in group to passive and hidden
      #   UU::OS::Cast.set_attributes('ues:TERRITORY:GROUP',
      #           :casted_subject_uri => 'ues:TERRITORY:ROLE',
      #           :state => 'PASSIVE',
      #           :visible => false
      #   )
      def set_attributes(subject_uri, cast = nil)
        svc = UU::OS::REST::RemoteClient.new(Cast)
        payload = UU::OS::Cast::CastSetAttributes.new(cast).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('setAttributes', subject_uri, payload)
        end
      end

      # Deletes the cast from the system. Any change of user authorization
      # resulting from the cast modification will take effect approximately after 15
      # minutes. The last active executive cast cannot be deleted when this role has an
      # active assistance or guest cast (active casts of these types cannot exist without
      # at least one active executive cast). The last active executive cast for this role cannot
      # be deleted when the role is competent for at least one artifact (a role without an
      # active cast cannot be competent for any artifacts).
      #
      # @param [String, UU::OS::UESURI] subject_uri UESURI of the subject to which the other is casted.
      # @param [CastDelete] cast DTO containing UESURI of the casted subject.
      # @option cast [String, UU::OS::UESURI] :casted_subject_uri UESURI of the subject (role or group) which is casted to the subject identified by the command main entity. Cannot be nil.
      #
      # @example
      #   # Delete a cast
      #   UU::OS::Cast.delete('ues:TERRITORY:ROLE2', :casted_subject_uri => 'ues:TERRITORY:ROLE1')
      def delete(subject_uri, cast = nil)
        svc = UU::OS::REST::RemoteClient.new(Cast)
        payload = UU::OS::Cast::CastDelete.new(cast).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('delete', subject_uri, payload)
        end
      end

      # Checks the cast existence of a subject (access role, role or group) to another subject (role or group) specified in subject_uri.
      # The cast subject is identified either by UESURI or by Universe ID of a user whose access role is cast to subjectUri.
      # Only the existence of a direct cast is checked by default. The existence of an indirect cast can be checked only between
      # an access role (identified either by UESURI or by Universe ID) and another subject (role or group).
      #
      # @param [String, UU::OS::UESURI] subject_uri UESURI of the role (role or group) for which the uuCommand checks whether another role (role, group or access role) is casted to.
      # @param [CastExists] cast Specific options.
      # @option cast [String, UU::OS::UESURI] :casted_subject_uri UESURI of the subject (role, group or access role) which has a cast to the subject_uri. Either casted_subject_uri or casted_subject_universe_id must be set.
      # @option cast [String] :casted_subject_universe_id Universe ID of a user whose access role is cast to the subjectUri. Only an active access role is taken into account. If the user has no active access role in the territory, then false is always returned. Either casted_subject_uri or casted_subject_universe_id must be set.
      # @option cast [Boolean] :recursive Flag indicates whether to also check an indirect cast. An indirect cast can be checked only between an access role (identified either by UESURI or by Universe ID) and another subject (role or group). The default value is false (only a direct cast is checked).
      #
      # @return [Boolean] True if cast exists, else false.
      #
      # @example
      #   # Check direct cast existence (the role has cast to another role or group).
      #   UU::OS::Cast.exists('ues:TERRITORY:GROUP', :casted_subject_uri => 'ues:TERRITORY:ROLE')
      #
      #   # Check indirect access role cast existence (the access role has cast to a role which has cast to another role).
      #   UU::OS::Cast.exists('ues:TERRITORY:ROLE', :casted_subject_uri => 'ues:TERRITORY:ACCESS_ROLE', :recursive => true)
      #
      #   # Check direct cast existence for user with selected UID (the user has an access role cast in another role or group).
      #   UU::OS::Cast.exists('ues:TERRITORY:ROLE', :casted_subject_universe_id => '1-1111-1')
      #
      #   # Check indirect cast existence for user with selected UID (the user has an access role cast to a role which is cast in another role).
      #   UU::OS::Cast.exists('ues:TERRITORY:ROLE', :casted_subject_universe_id => '1-1111-1', :recursive => true)

      def exists(subject_uri, cast = nil)
        svc = UU::OS::REST::RemoteClient.new(Cast)
        dto = UU::OS::Cast::CastExists.new(cast)
        svc.add_parameter('castedSubjectUri', dto.casted_subject_uri)
        svc.add_parameter('castedSubjectUniverseId', dto.casted_subject_universe_id)
        svc.add_parameter('recursive', dto.recursive)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('exists', subject_uri)
          return (res == 'true') || (res == '"true"')
        end
      end

      # Verifies the cast existence of a user to a list of subjects in a given territory. The user is identified by either Universe ID or access role UESURI.
      #
      # @param [String, UU::OS::UESURI] territory_uri UESURI of the territory where the subjects are located in.
      # @param [CastVerifyCastExistence] Object containing one of the user identifiers and a list of subjects.
      # @option criteria [String, UU::OS::UESURI] :cast_subject_uri UESURI of the user access role whose cast existence to subject_uris is verified. Either cast_subject_universe_id or cast_subject_uri must be set.
      # @option criteria [String] :cast_subject_universe_id Universe ID of the user whose cast existence to subject_uris is verified. The user must have an active access role in the territory. If the user does not exist or has no active access role in the territory, then an empty array is returned. Either cast_subject_universe_id or cast_subject_uri must be set.
      # @option criteria [UU::OS::UESURI[]] :subject_uris List of subjects in which the user cast existence is verified. The maximum number of subjects is 10. The subjects existence is not checked (an exception is NOT thrown if some of the subjects do not exist). If the user access role is included in subject_uris, then it will not be returned when verifying by the user access role itself (by cast_subject_uri), but it will be returned when verifying by the user Universe ID (by cast_subject_universe_id).
      #
      # @return [UU::OS::UESURI[]] - List of subjects with an existing cast. UESURI is in the normalized form (with both codes and OIDs).
      #
      # @example
      #   # Gets subjects with an existing cast of an Access Role specified by UID in a given territory.
      #   UU::OS::Cast.verify_cast_existence("ues:UU-BT:UU-BT",
      #     :cast_subject_universe_id => "6-6666-1",
      #     :subject_uris => [
      #       "ues:[902387469238]:[987236489734]",
      #       "ues:[902387469238]:[967698769878]",
      #       "ues:[902387469238]:[998762378232]"
      #   ])
      #
      #   # Gets subjects with an existing cast of an Access Role specified by UESURI in a given territory.
      #   UU::OS::Cast.verify_cast_existence("ues:UU-BT:UU-BT",
      #   :cast_subject_uri => "ues:UU-BT:6-6666-1",
      #   :subject_uris => [
      #      "ues:[902387469238]:[987236489734]",
      #      "ues:[902387469238]:[967698769878]",
      #      "ues:[902387469238]:[998762378232]"
      #   ])
      
      def verify_cast_existence(territoryUri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(Cast)
        dto = UU::OS::Cast::CastVerifyCastExistence.new(criteria)
        svc.add_parameter('castSubjectUri', dto.cast_subject_uri)
        svc.add_parameter('castSubjectUniverseId', dto.cast_subject_universe_id)
        svc.add_parameter('subjectUris', dto.subject_uris)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('verifyCastExistence', territoryUri)
          return JSON.parse(res)
        end
      end

      # Returns a cast history list in the subject and/or from the subject. Only direct casts are supported. Casts are filtered and ordered by the query criteria. By default, only casts IN the subject are returned, sorted by the castCreationStamp. Result list cannot be paged. The maximum number of casts returned is 10,000. In case of exceeding the limit, an exception is thrown. Cast strategy 1 or higher must be enabled for the territory to use this command.
      #
      # @param [String] subject_uri The subject where the subjects are cast, identified by UESURI.
      # @param [CastGetCastHistoryList] criteria Object containing the search criteria.
      # @option criteria [String] :query Query in uuQuery 2.0 format. Records can be filtered by all attributes of {Cast::CastHistoryAttributes CastHistoryAttributes}. If filtering by an UESURI, it must be in the canonical form ues:[TER_OID]:[ROLE_OID]. If set to nil, no filtering will be applied except for {Cast::CastDirection castDirection} which will be set to {Cast::CastDirection IN} value by default.
      #
      # @return [Array] Array of {Cast::CastHistoryAttributes CastHistoryAttributes} containing the cast history attributes. The list is sorted by the cast creation stamp.
      #
      # @example
      #   UU::OS::Cast.get_cast_history_list('ues:TERRITORY:ROLE_OR_GROUP',
      #         :query => "castSubjectName='Name' AND castSubjectCode='Code' ORDER BY castSubjectName")
      def get_cast_history_list(subject_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(Cast)
        dto = Cast::CastGetCastHistoryList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getCastHistoryList', subject_uri)
          return UU::OS::REST::ResultList.new(Cast::CastGetCastHistoryList, Cast::CastHistoryAttributes, res)
        end
      end

    end
  end

end
