require 'uu/os/rest'
require 'uu/os/qos/qos_handler'
require 'uu/os/command/command_attributes'
require 'uu/os/command/command_create'
require 'uu/os/command/command_set_attributes'
require 'uu/os/command/command_export'
require 'uu/os/command/command_get_export_data'

# Module Unicorn Universe.
module UU
  # Module Operating System.
  module OS
    # Module Command.
    module Command
      extend self

      # Service path
      PATH = 'uu/os/Command'
      # The command creates a command in a specified location. At the very least meta artifact must be specified in
      # {CommandCreate} Object. Competent role for the new command will be selected as the most suitable according
      # to specified container and meta artifact (executive/authorized role with connected interface), or can be also specified in
      # {CommandCreate} Object.
      #
      # @param [String, UU::OS::UESURI] location_uri URI of the uuApp Package where the command should be created.
      # @param [CommandCreate] criteria Object containing attributes for the new command.
      # @option criteria [String] :name Name of the new command. If nil is set, default name from meta artifact is used if set, otherwise the create command fails.
      # @option criteria [String] :code Code of the new command. If nil is set, the code is generated automatically using the system sequence. The code must be unique in the given territory, otherwise the create command fails.
      # @option criteria [String] :description Description of the new command. If nil is set, default description from meta artifact is used if set, otherwise no description is set on the new command.
      # @option criteria [String, UU::OS::Artifact::SecurityLevel] :security_level Security level. If nil is set, security level is taken from the meta artifact default security level. If the meta artifact template security level is higher than the user's security clearance, the create command fails.
      # @option criteria [String, UU::OS::UESURI] :icon_uri URI of the command's icon. If nil is set, the role uses icon from meta artifact. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option criteria [String] :controller_url_path Path at which the controller will be deployed. Value must start with '/' and contain only valid URL characters.
      # @option criteria [String] :asynchronous Flag if command is asynchronous. Defaults to false (command is created as synchronous by default).
      # @option criteria [String, UU::OS::UESURI] :competent_role_uri URI of the competent role. If nil is set, the executive role from authorization is used. If no role is authorized to create the command into specified container, the create command fails. If the selected role doesn't have connected the role interface with the selected meta artifact, the create command fails. If the selected role doesn't belong to the same organizational unit as a new command or if it doesn't belong to one of superior organizational units, the create command fails.
      # @option criteria [String, UU::OS::UESURI] :meta_artifact_uri URI of the meta artifact. Only a meta artifact with the template class of Command (not its descendants) can be used. If the rights does not permit to create a new command from the meta artifact in the specified location, the create command fails.
      #
      # @return [UU::OS::UESURI] URI of the created command.
      #
      # @example
      #   # Create command
      #   cmd_uri = UU::OS::Command.create('ues:TERRITORY:APP_PACKAGE', :name => 'New Command', :code => 'NEW_COMMAND',
      #             :controller_url_path => '/ObjectType/command', :meta_artifact_uri => 'ues:TERRITORY:META_ARTIFACT',
      #             :asynchronous => false)
      def create(location_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(Command)
        payload = UU::OS::Command::CommandCreate.new(criteria).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('create', location_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # The command deletes the specified command from the system. The command fails when the specified command is
      # referenced by some use case.
      #
      # @param [String, UU::OS::UESURI] uri URI of the command to be deleted.
      #
      # @return [void]
      #
      # @example
      #   # Delete command
      #   UU::OS::Command.delete('ues:TERRITORY:COMMAND')
      def delete(uri)
        svc = UU::OS::REST::RemoteClient.new(Command)
        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('delete', uri)
        end
      end

      # This command returns attributes representing the command specified by the command_uri parameter.
      #
      # @param [String, UU::OS::UESURI] command_uri URI of the command.
      #
      # @return [CommandAttributes] Object containing the command attributes.
      #
      # @example
      #   # Get command attributes.
      #   attributes = UU::OS::Command.get_attributes('ues:TERRITORY:COMMAND')
      def get_attributes(command_uri)
        svc = UU::OS::REST::RemoteClient.new(Command)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttributes', command_uri)
          return Command::CommandAttributes.new(res)
        end
      end

      # The command for setting attributes of a specified command. The command can't modify references to objects like storage,
      # competent role and so on.
      #
      # @param [String, UU::OS::UESURI] uri URI of the command.
      # @param [CommandSetAttributes] criteria Object containing new attributes for given command.
      # @option criteria [String] :name The new name of the command. If nil is set, the name is not modified.
      # @option criteria [String] :code The new code of the command. If nil is set, the code is not modified.
      # @option criteria [String] :description The new description of the command. If nil is set, the description is not modified.
      # @option criteria [String, UU::OS::Artifact::SecurityLevel] :security_level Security level of the command. If nil is set, the security level is not modified.
      # @option criteria [String, UU::OS::UESURI] :icon_uri URI of the command icon. If nil is set, the icon is not modified. If nil URI ("ues:{[-1]}:{[-1]}:") is set, the command icon is removed and the command uses icon from meta artifact. An example of icon URI: 'ues:#{{}system}:#{{}ues_v5.core_v1.codetable_v1.base_v1.CodeTable_Icons}:#{{}ART_077}'.
      # @option criteria [String] :controller_url_path Path at which the controller will be deployed. If nil is set, the controller_url_path is not modified. Value must start with '/' and contain only valid URL characters.
      # @option criteria [String] :asynchronous Flag if command is asynchronous. If nil is set, the asynchronous flag is not modified.
      #
      # @return [UU::OS::UESURI] URI of the changed command
      #
      # @example
      #   # Set command attributes.
      #   cmd_uri = UU::OS::Command.set_attributes('ues:TERRITORY:COMMAND', :description => 'Command designated for testing purposes',
      #     :name => 'New Command', :asynchronous => true)
      def set_attributes(uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(Command)
        payload = UU::OS::Command::CommandSetAttributes.new(criteria).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('setAttributes', uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Exports a command to the XML file, which is saved to the export storage. The export is an asynchronous process.<br>
      # It is possible to wait for the end of the process. More information in {UU::OS::Env::Process Process} and in {UU::OS::REST::Future Future}.
      #
      # @param [String, UU::OS::UESURI] uri URI of the command to export.
      #
      # @return [UU::OS::UESURI] URI of the export process.
      #
      # @example
      #   # Start exporting a command.
      #   process_uri = UU::OS::Command.export('ues:TERRITORY:COMMAND')
      #
      #   # Wait for the end of the process.
      #   export_uri = UU::OS::REST::Future.new(process_uri).get(nil, 600)
      def export(uri)
        svc = UU::OS::REST::RemoteClient.new(Command)
        payload = UU::OS::Command::CommandExport.new(nil).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('export', uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end

      # Returns a XML file generated by the export command {UU::OS::Command#export export}.
      #
      # @param [String, UU::OS::UESURI] command_uri URI of the command we want to get the exported XML file from.
      # @param [UU::OS::Command::CommandGetExportData] command Specific options.
      # @option command [String] :requested_mime_type MIME type of the returned value. All possible MIME types can be found as constants in {UU::OS::Artifact::ExportMimeType ExportMimeType} class. ZIP is returned by default.
      #
      # @return [UU::OS::REST::BinaryValue] Data file of an exported command (XML or ZIP containing XML).
      #
      # @example
      #   # Start exporting a command
      #   process_uri = UU::OS::Command.export('ues:TERRITORY:COMMAND')
      #
      #   # Wait for the end of the process.
      #   export_uri = UU::OS::REST::Future.new(process_uri).get(nil, 600)
      #
      #   # Get exported XML file.
      #   exported_data = UU::OS::Command.get_export_data(export_uri, :requested_mime_type => 'application/xml')
      def get_export_data(command_uri, command = nil)
        svc = UU::OS::REST::RemoteClient.new(Command)
        dto = UU::OS::Command::CommandGetExportData.new(command)
        if dto.requested_mime_type
          svc.add_parameter('requestedMimeType', dto.requested_mime_type)
        end
        UU::OS::QoS::QoSHandler.auto_retry do
          return svc.get_binary('getExportData', command_uri)
        end
      end

    end

  end
end
