require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/sheet/mime_type'
require 'uu/os/rest/binary_value'
require 'uu/os/sheet_template/sheet_template_attributes'
require 'uu/os/sheet_template/sheet_template_check_in'
require 'uu/os/sheet_template/sheet_template_create'
require 'uu/os/sheet_template/sheet_template_get_data'
require 'uu/os/sheet_template/sheet_template_set_attributes'
require 'uu/os/sheet_template/sheet_template_get_sheet_template_list'

module UU
  module OS
    module SheetTemplate
      extend self
      # Service path
      PATH = 'ues/core/content/UESSheetTemplate'
      
      # Returns attributes representing specified sheet template.
      #
      # @param [String, UU::OS::UESURI] sheet_template_uri URI of sheet template which attributes will be retrieved
      #      
      # @return [SheetTemplateAttributes] Object containing the sheet template attributes
      # @example      
      #
      #   #Get attributes of a sheet template
      #   UU::OS::SheetTemplate.get_attributes('ues:TERRITORY:META_ARTIFACT:SHEET_TEMPLATE')
      def get_attributes(sheet_template_uri)
        svc = UU::OS::REST::RemoteClient.new(SheetTemplate)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getAttributes', sheet_template_uri)
          return SheetTemplate::SheetTemplateAttributes.new(res)
        end
      end
      
      # Replaces the content of the specified sheet template and unlock it by default. Sheet template content must be in UXML or UDF format.
      # Sheet template must be locked by the current user or must be unlocked.
      #
      # @param [String, UU::OS::UESURI] sheet_template_uri URI of the sheet template.
      # @param [SheetTemplateCheckIn] sheet_template Attributes of importing sheet template.
      # @option sheet_template [String] :name The new name of the sheet template. When null is set, the name is not modified. The name must be unique in a given Meta Artifact, otherwise the command fails.
      # @option sheet_template [String] :code The new code of the sheet template. When null is set, the code is not modified. The code must be unique in a given Meta Artifact, otherwise the command fails.
      # @option sheet_template [String] :description The new description of the sheet template. When null is set, the description is not modified.
      # @option sheet_template [File, UU::OS::REST::BinaryValue] :content Not nil. The new content of the sheet template in UXML or UDF format.
      # @option sheet_template [String] :version_label The version label of the new sheet template version. When null is set, it is left empty.
      # @option sheet_template [String, UU::OS::Env::VersionStrategy] :version_strategy Specifies how to work with versions (see {UU::OS::Env::VersionStrategy} for possible values). When null is set, the OVERWRITE_VERSION strategy is used.
      #
      # @example      
      #   #Create data for check in
      #   data = UU::OS::BinaryValue.new(File.open('document.xml', 'rb'))
      #   
      #   #Check in sheet template
      #   UU::OS::SheetTemplate.check_in('ues:TERRITORY:META_ARTIFACT:SHEET_TEMPLATE', :name => 'newName', :code => 'newCode', :description => 'newDesc',
      #          :content => data, :version_label => 'newLabel', :version_strategy => 'OVERWRITE_VERSION') 
      def check_in(sheet_template_uri, sheet_template = nil)
        svc = UU::OS::REST::RemoteClient.new(SheetTemplate)
        dto = UU::OS::SheetTemplate::SheetTemplateCheckIn.new(sheet_template)
        payload = dto.to_hash
        payload.delete(:content)
        tmpfile = Tempfile.new("SheetTemplateCheckIn")
        tmpfile.binmode
        begin
          if (!dto.content.kind_of?(UU::OS::REST::BinaryValue))
            dto.content = UU::OS::REST::BinaryValue.new(dto.content)
          end
          if (dto.content.data.respond_to?(:read))
            if (dto.content.data.respond_to?(:path))
              payload[:content] = dto.content.to_hash(false)
            else
              dto.content.data.rewind
              tmpfile.write(dto.content.data.read)
              payload[:content] = tmpfile
            end
          else
            tmpfile.write(dto.content.data)
            payload[:content] = tmpfile
          end
          tmpfile.flush
          tmpfile.rewind
  
          UU::OS::QoS::QoSHandler.auto_retry do
            svc.post('checkIn', sheet_template_uri, payload)
          end
        ensure
          tmpfile.close
          tmpfile.unlink
        end
      end
      
      # Returns the content of the sheet template. The content may be returned as structured UXML or full content including images in a
      # compressed UDF format. The type of the returned stream is determined by options (see {SheetTemplateGetData#requested_mime_type}),
      # UXML is returned by default.
      #
      # @param [String, UU::OS::UESURI] sheet_template_uri URI of the sheet template.
      # @param [SheetTemplateGetData] sheet_template Object containing options for content retrieving
      # @option sheet_template [String, UU::OS::Sheet::MimeType] :requested_mime_type The format of the returned sheet template content. Possible values are {UU::OS::Sheet::MimeType#UXML} and {UU::OS::Sheet::MimeType#UARCHIVE}. When null is set, the <code>UXML</code> is used.
      #
      # @return [UU::OS::REST::BinaryValue] sheet template content.
      # @example          
      #   #Get sheet template data
      #   UU::OS::SheetTemplate.get_data('ues:TERRITORY:META_ARTIFACT:SHEET_TEMPLATE', :requested_mime_type => UU::OS::Sheet::MimeType::UXML)
      def get_data(sheet_template_uri, sheet_template = nil)
        svc = UU::OS::REST::RemoteClient.new(SheetTemplate)
        dto = UU::OS::SheetTemplate::SheetTemplateGetData.new(sheet_template)
        svc.add_parameter('requestedMimeType', dto.requested_mime_type)

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get_binary('getData', sheet_template_uri)
          if res
            if (dto.requested_mime_type == UU::OS::Sheet::MimeType::UXML)
              return UU::OS::REST::BinaryValue.new(StringIO.new(res.data.read.force_encoding("UTF-8")),false)
            else
              return res
            end
          else
            return nil
          end
        end
      end
      
      # Returns a list of sheet templates on the specified Meta Artifact. Command does not return all sheet templates, only a sublist. The list
      # of returned sheet templates could be filtered by the name, code, main, visible, order of a sheet template, state of the lock and width.
      # The list is sorted by the name (and code in case are equal) of a sheet template by default, or it can be sorted arbitrarily by any of
      # sheet template attributes usable for the filtering or these attributes combinations. See the UES query documentation for more
      # information.
      #
      # @param [String, UU::OS::UESURI] meta_artifact_uri URI of Meta Artifact.
      # @param [SheetTemplateGetSheetTemplateList] criteria Specific options.
      # @option criteria [String] :query Query for filtering and ordering of the result list (more information in {uuQuery - Guideline}[https://plus4u.net/ues/sesm?SessFree=ues%253AUNI-BT%255B210795%255D%253AUES.DOC%252FCMD_V1%252FUESQUERY%255B44191574519623267%255D%253A]).
      # It is possible to filter result by: {SheetTemplateAttributes#name name},
      # {SheetTemplateAttributes#code code}, {SheetTemplateAttributes#order order},
      # {SheetTemplateAttributes#main main}, {SheetTemplateAttributes#locked locked},
      # {SheetTemplateAttributes#lock_time lockTime}, {SheetTemplateAttributes#visible visible} and
      # {SheetTemplateAttributes#width width} : {UU::OS::Sheet::SheetWidthType}.
      #
      # @return [Array] An Array with the command result. Contains instances of {UU::OS::SheetTemplate::SheetTemplateAttributes}.
      # @example        
      #   #Get sheet template list
      #   UU::OS::SheetTemplate.get_sheet_template_list('ues:TERRITORY:META_ARTIFACT', :query => "name='Description%'")
      def get_sheet_template_list(meta_artifact_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(SheetTemplate)
        dto = SheetTemplate::SheetTemplateGetSheetTemplateList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getSheetTemplateList', meta_artifact_uri)
          return UU::OS::REST::ResultList.new(SheetTemplate::SheetTemplateGetSheetTemplateList, SheetTemplate::SheetTemplateAttributes, res)
        end
      end
      
      # Creates a new sheet template on the specified Meta Artifact. Initial sheet template content can be specified.
      #
      # @param [String, UU::OS::UESURI] meta_artifact_uri URI of Meta Artifact.
      # @param [SheetTemplateCreate] sheet_template Object containing attributes of the sheet template.
      # @option sheet_template [String] :name Name of the new sheet template. When null is set, the name is generated automatically. The name must be unique in a given Meta Artifact, otherwise the command fails.
      # @option sheet_template [String] :code Code of the new sheet template. When null is set, the code is generated automatically using the system sequence. The code must be unique in a given Meta Artifact, otherwise the command fails.
      # @option sheet_template [String] :description Description of the new sheet template.
      # @option sheet_template [UU::OS::UESURI] :source_template_uri UESURI of the sheet template from which the new sheet template will be copied. Source sheet template must be located on the Meta Artifact from which the Meta Artifact with new sheet template was copied.
      # @option sheet_template [Boolean] :main Specifies whether created sheet template is set as the main sheet template on Meta Artifact. When not set, sheet template is not set as main. When true is set, the sheet template should be visible.
      # @option sheet_template [Integer] :order Specifies the order (or index) of the new sheet template on the specified Meta Artifact. All sheet templates on a Meta Artifact are
      #     ordered sequentially. The first sheet template has index 0. When null is set or when the order exceeds the number of sheet templates on
      #     the Meta Artifact, the sheet template is inserted at the end of the sheet template list. When 0 or less is set, the sheet template is
      #     inserted at the beginning of the sheet template list. Sheet template currently at that position (if any) and any subsequent sheet
      #     templates are shifted to the right (adds one to their order).
      # @option sheet_template [File, UU::OS::REST::BinaryValue] :content Initial content of sheet template. Must be in UXML or UDF format.
      # @option sheet_template [String] :version_label Label of the new sheet template version.
      # @option sheet_template [Boolean] :visible Specifies the visibility of the sheet template.
      # @option sheet_template [String, UU::OS::Sheet::SheetWidthType] :width Width of the sheet template.
      # @option sheet_template [Boolean] :implicit_creation Specifies whether sheet should be created by the sheet template simultaneously with artifact creation.
      #
      # @return [UU::OS::UESURI] UESURI of created sheet template.
      # @example 
      #   #Create initial data
      #   data = UU::OS::BinaryValue.new(File.open('document.xml', 'rb'))
      #
      #   #Create sheet template
      #   UU::OS::SheetTemplate.create('ues:TERRITORY:META_ARTIFACT', :name => 'testTemplate', :code => 'TEMPLATE_CODE', :description => 'newDesc', :content => data,
      #         :main => false, :order => 0, :version_label => 'New Sheet Template', :visible => true, :implicit_creation => false,
      #         :width => UU::OS::Sheet::SheetWidthType::WIDTH_1226PX) 
      def create(meta_artifact_uri, sheet_template = nil)
        svc = UU::OS::REST::RemoteClient.new(SheetTemplate)
        dto = UU::OS::SheetTemplate::SheetTemplateCreate.new(sheet_template)
        
        payload = dto.to_hash
        payload.delete(:content)
        tmpfile = Tempfile.new("SheetTemplateCreate")
        tmpfile.binmode
        begin
          if (!dto.content.nil?)
            if (!dto.content.kind_of?(UU::OS::REST::BinaryValue))
              dto.content = UU::OS::REST::BinaryValue.new(dto.content)
            end
            if (dto.content.data.respond_to?(:read))
              if (dto.content.data.respond_to?(:path))
                payload[:content] = dto.content.to_hash(false)
              else
                dto.content.data.rewind
                tmpfile.write(dto.content.data.read)
                payload[:content] = tmpfile
              end
            else
              tmpfile.write(dto.content.data)
              payload[:content] = tmpfile
            end
          end
          tmpfile.flush
          tmpfile.rewind
        
          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('create', meta_artifact_uri, payload)
            return UU::OS::UESURI.new(res)
          end
        ensure
          tmpfile.close
          tmpfile.unlink
        end
      end
      
      # Deletes the sheet template from the system.
      #
      # @param [String, UU::OS::UESURI] sheet_template_uri UESURI of the sheet template which is about to be deleted.
      #
      # @example          
      #   #Delete sheet template
      #   UU::OS::SheetTemplate.delete('ues:TERRITORY:META_ARTIFACT:SHEET_TEMPLATE')
      def delete(sheet_template_uri)
        svc = UU::OS::REST::RemoteClient.new(SheetTemplate)
        UU::OS::QoS::QoSHandler.auto_retry do
          svc.post('delete', sheet_template_uri)
        end
      end
      
      # Updates basic attributes of the sheet template.
      #
      # @param [String, UU::OS::UESURI] sheet_template_uri UESURI of the sheet template.
      # @param [SheetTemplateSetAttributes] sheet_template Object containing the sheet template attributes.
      # @option sheet_template [String] :name The new name of the sheet template. When null is set, the name is not modified. The name must be unique in a given Meta Artifact, otherwise the command fails.
      # @option sheet_template [String] :code The new code of the sheet template. When null is set, the code is not modified. The code must be unique in a given Meta Artifact, otherwise the command fails.
      # @option sheet_template [Boolean] :main Specifies if the sheet template is to be set as main on the Meta Artifact. When null is set, the main is not modified. When true is set, the
      #     sheet template must be visible. When false is set and the sheet template is set as main, system sets the first visible sheet template as main. If
      #     the main sheet template is at the same time the first visible, nothing happens.
      # @option sheet_template [Boolean] :visible Specifies if the sheet template is set as visible. When null is set, the visible is not modified. When false is set, the sheet template cannot be main at the same time.
      # @option sheet_template [Integer] :order Specifies the order (or index) of the sheet template on the specified Meta Artifact. All sheet templates are ordered sequentially. The first
      #     sheet template has index 0. When 0 or less is set, the sheet template is inserted at the beginning of the sheet template list. When null is set,
      #     the order is not modified. Sheet template currently at that position (if any) and any subsequent sheet templates are shifted to the right (adds
      #     one to their order).
      # @option sheet_template [String] :description The new description of the sheet template. When null is set, the description is not modified.
      # @option sheet_template [String, UU::OS::Sheet::SheetWidthType] :width The new width of the sheet template. When null is set, the width is not modified.
      # @option sheet_template [Boolean] :implicit_creation Specifies whether sheet should be created by the sheet template simultaneously with artifact creation.
      #
      # @return [UU::OS::UESURI] UESURI of the changed sheet template.
      #
      # @example          
      #   #Set attributes of sheet template
      #   UU::OS::SheetTemplate.set_attributes('ues:TERRITORY:META_ARTIFACT:SHEET_TEMPLATE', :name => 'newName', :code => 'newCode', :main => true, :visible => true, 
      #       :order => 3, :description => 'newDesc', :width => UU::OS::Sheet::SheetWidthType::WIDTH_650PX, :implicit_creation => true)
      def set_attributes(sheet_template_uri, sheet_template = nil)
        svc = UU::OS::REST::RemoteClient.new(SheetTemplate)
        payload = UU::OS::SheetTemplate::SheetTemplateSetAttributes.new(sheet_template).to_json

        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.post('setAttributes', sheet_template_uri, payload)
          return UU::OS::UESURI.new(res)
        end
      end
      
    end
  end
end